#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

void try(const char * const what, int result) {
    if (!result) return;
    int errno_local = errno;

    printf("\n");
    fflush(stdout);
    fprintf(stderr,"error: %s failed\n",what);
    if (errno_local != 0)
    {
        errno = errno_local;
        perror(NULL);
    }
    exit(1);
}

unsigned int read32(unsigned char * buf) {
    return (buf[0]<<24)|(buf[1]<<16)|(buf[2]<<8)|(buf[3]);
}
unsigned int read16(unsigned char * buf) {
    return (buf[0]<<8)|(buf[1]);
}

void hexprint(const unsigned char * bytes, int len) {
    int i;
    for (i=0;i<len;i++) {
        if (i>0 && i%16 == 0) printf("\n");
        printf("%02x ",bytes[i]);
    }
    printf("\n");
}

enum {DSPSIZE=144};
/* BUFSIZ is at least 256, so a DSPSIZE header will fit */
unsigned char buf[BUFSIZ];
const char padding[0x40];

void write_file(FILE *infile, FILE *outfile, long total_bytes, long padding_bytes)
{
    long bytes_left;
    for ( bytes_left = total_bytes; bytes_left >= BUFSIZ; bytes_left -= BUFSIZ)
    {
        try("read data",fread(buf,1,BUFSIZ,infile)!=BUFSIZ);
        try("write data",fwrite(buf,1,BUFSIZ,outfile)!=BUFSIZ);
    }
    if ( bytes_left > 0 )
    {
        try("read data",fread(buf,1,bytes_left,infile)!=bytes_left);
        try("write data",fwrite(buf,1,bytes_left,outfile)!=bytes_left);
    }
    if ( padding_bytes != 0)
    {
        try("write padding",fwrite(padding,1,padding_bytes,outfile)!=padding_bytes);
    }
}

int main(int argc, char ** argv) {
    int block_RND;
    int block_SSD;
    int block_SRT;
    int block_DSP;
    int stream_count;
    int i;
    int channels = 1;


    FILE * srtfile = NULL;
    long srt_length;
    FILE * dspfile = NULL;
    long dsp_length;
    FILE * dsp2file = NULL;
    long dsp2_length;

    FILE * ssdfile;
    int toreplace;
    long dspoffset;
    long dsp_data_size;

    if (argc!=5 && argc!=6) {printf("usage: %s blah.srt 2 blah.ssd blah02.dsp [blah02R.dsp]\n",argv[0]); return 1;}

    if (argc == 6)
    {
        channels = 2;
    }
    else
    {
        printf("mono is yet unknown\n");
        exit(1);
    }

    try("open srt",(srtfile=fopen(argv[1],"r+b"))==NULL);
    try("seek to srt file end",fseek(srtfile,0,SEEK_END));
    try("get srt file size",(srt_length=ftell(srtfile))==-1);
    try("get back to srt file start",fseek(srtfile,0,SEEK_SET));

    try("open dsp file",(dspfile=fopen(argv[4],"rb"))==NULL);
    try("seek to dsp file end",fseek(dspfile,0,SEEK_END));
    try("get dsp file size",(dsp_length=ftell(dspfile))==-1);
    try("seek back to dsp file start",fseek(dspfile,0,SEEK_SET));

    if (channels == 2)
    {
        try("open second dsp file",(dsp2file=fopen(argv[5],"rb"))==NULL);
        try("seek to second dsp file end",fseek(dsp2file,0,SEEK_END));
        try("get second dsp file size",(dsp2_length=ftell(dsp2file))==-1);
        try("seek back to second dsp file start",fseek(dsp2file,0,SEEK_SET));
    }

    try("open ssd",(ssdfile=fopen(argv[3],"wb"))==NULL);

    try("read srt header",fread(buf,1,16,srtfile)!=16);

    block_RND=read32(buf);
    block_SSD=read32(buf+4);
    block_SRT=read32(buf+8);
    block_DSP=read32(buf+12);

    try("parse stream # to replace",sscanf(argv[2],"%d",&toreplace)!=1);

    printf("%s",argv[4]);
    if (channels == 2)
    {
        printf(",%s",argv[5]);
    }
    printf("->%s[%d]+%s...",argv[1],toreplace,argv[3]);

    try("seek to SSD block",fseek(srtfile,block_SSD,SEEK_SET));
    try("read SSD block header",fread(buf,1,16,srtfile)!=16);
    try("check SSD block header",memcmp("SSD\0",buf,4)||memcmp("\0\0\0\x10\0\0\0\0",buf+8,8));

    /* NOTE: only stereo here on out */

#if 0
    stream_count = read32(buf+4);

    if (toreplace<0 || toreplace>=stream_count) {
        printf("bad stream number, there are %d stream headers in this .srt\n",stream_count);
        exit(1);
    }
#endif

    try("seek to DSP block",fseek(srtfile,block_DSP,SEEK_SET));
    try("read DSP block header",fread(buf,1,16,srtfile)!=16);
    try("check DSP block header",memcmp("DSP\0",buf,4)||memcmp("\0\0\0\x10\0\0\0\0",buf+8,8));

    /* check stream count */
    stream_count = read32(buf+4);

    if (toreplace<0 || toreplace>=stream_count) {
        printf("bad stream number, there are %d stream headers in this .srt\n",stream_count);
        exit(1);
    }

    try("seek to DSP entry index",fseek(srtfile,block_DSP+0x10+toreplace*4,SEEK_SET));
    try("read DSP entry index",fread(buf,1,4,srtfile)!=4);
    dspoffset = read32(buf);

    /* check DSP size */
    {
        long next_offset = srt_length - (block_DSP+0x10);
        if (toreplace < (stream_count-1))
        {
            try("seek to next DSP entry index",fseek(srtfile,block_DSP+0x10+(toreplace+1)*4,SEEK_SET));
            try("read next DSP entry index",fread(buf,1,4,srtfile)!=4);
            next_offset = read32(buf);
        }

        if (next_offset - dspoffset != DSPSIZE)
        {
            printf("DSP table has unexepcted entry size %ld\n",
                    next_offset - dspoffset);
            exit(1);
        }
    }

    /* build DSP header */
    {
        char newhead[DSPSIZE] = {0,0,0,0, 0,0,0,1, 0,0,0,0, 1,0,0,0};

        try("read input dsp header",fread(buf,1,0x60,dspfile)!=0x60);

        /* copy sample rate */
        memcpy(&newhead[0x08], &buf[0x08], 4);

        /* FIRST CHANNEL */

        /* copy loop start/end */
        memcpy(&newhead[0x10], &buf[0x10], 8);

        /* copy data size */
        memcpy(&newhead[0x18], &buf[0x04], 4);

        /* copy initial nibble addr */
        memcpy(&newhead[0x1C], &buf[0x18], 4);

        /* copy channel codec setup */
        memcpy(&newhead[0x20], &buf[0x1C], 0x30);

        dsp_data_size = read32(&buf[0x04]);

        try("read 2nd input dsp header",fread(buf,1,0x60,dsp2file)!=0x60);

        /* SECOND CHANNEL */

        /* check sample rate */
        try("check 2nd channel srate",memcmp(&newhead[0x08], &buf[0x08], 4));

        /* copy loop start/end */
        memcpy(&newhead[0x50], &buf[0x10], 8);

        /* copy data size */
        memcpy(&newhead[0x58], &buf[0x04], 4);

        /* copy initial nibble addr */
        memcpy(&newhead[0x5C], &buf[0x18], 4);

        /* copy codec setup */
        memcpy(&newhead[0x60], &buf[0x1C], 0x30);

        try("check for matching sizes", dsp_data_size != read32(&buf[0x04]));

        /* write new header */
        try("seek to DSP entry",fseek(srtfile,block_DSP+0x10+dspoffset,SEEK_SET));
        try("replace DSP entry",fwrite(newhead,1,DSPSIZE,srtfile)!=DSPSIZE);
    }

    /* write data */
    {
        /* round nibble count up to 64 bytes */
        const long total_bytes = (dsp_data_size + 0x7F) / 0x80 * 0x40;
        const long padding_bytes = total_bytes - ((dsp_data_size + 1) / 2);
        const long total_real_bytes = total_bytes - padding_bytes;

        write_file(dspfile,ssdfile,total_real_bytes,padding_bytes);
        write_file(dsp2file,ssdfile,total_real_bytes,padding_bytes);
    }

    /* cleanup */
    try("close srt",EOF==fclose(srtfile));
    fclose(dspfile);
    if (dsp2file)
    {
        fclose(dsp2file);
    }
    try("close ssd",EOF==fclose(ssdfile));

    printf(" done!\n");

    return 0;
}
